/*
 * Copyright (C) 2001,2002 Wojciech Purczynski <cliph@isec.pl>
 *
 * THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
 * IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
 * THE FREE SOFTWARE  FOUNDATION;  EITHER  VERSION 2 OF THE LICENSE, OR
 * (AT YOUR OPTION) ANY LATER VERSION.
 *
 * THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 * BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED  WARRANTY OF
 * MERCHANTABILITY OR FITNESS  FOR A PARTICULAR  PURPOSE.  SEE THE
 * GNU GENERAL PUBLIC LICENSE FOR MORE DETAILS.
 */

#include "module.h"

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/dcache.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <asm/namei.h>

#include "config.h"
#include "fs.h"

/* modified version of d_path() */
char *resolve_path(struct dentry *dentry, char *buffer, int buflen)
{
	char * end = buffer+buflen;
	char * retval;

	*--end = '\0';
	buflen--;
	if (dentry->d_parent != dentry && list_empty(&dentry->d_hash)) {
		buflen -= 10;
		end -= 10;
		memcpy(end, " (deleted)", 10);
	}

	/* Get '/' to the right */
	retval = end-1;
	*retval = '/';

	for (;;) {
		struct dentry * parent;
		int namelen;

		dentry = dentry->d_covers;
		parent = dentry->d_parent;
		if (dentry == parent)
			break;
		namelen = dentry->d_name.len;
		buflen -= namelen + 1;
		if (buflen < 0) {
			retval=ERR_PTR(-ENAMETOOLONG);
			break;
		}
		end -= namelen;
		memcpy(end, dentry->d_name.name, namelen);
		*--end = '/';
		retval = end;
		dentry = parent;
	}
	return retval;
} /* resolve_path */

/* NOTE: filename is from user-space memory */
char *get_realpath(const char * filename)
{
	char * buf, * res;
	struct dentry * dentry;

	dentry=namei(filename);
	if (IS_ERR(dentry)) return (char*)dentry;
		
	buf=__getname();
	if (IS_ERR(buf)) return buf;
	
	res=resolve_path(dentry, buf, PAGE_SIZE);
	if (IS_ERR(res)) {
		putname(buf);
		printk(KERN_ERR
		       NAME ": Unable to resolve realpath name");
	}
	
	dput(dentry);
	return res;
}

/*
 * Name got with get_exename must be released with put_exename()
 */
char *get_exename(void)
{
	struct vm_area_struct *vma;
	struct dentry * entry=NULL;
	char *buf, *res;

	vma=current->mm->mmap;
	while (vma) {
		if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file) {
			entry=vma->vm_file->f_dentry;
			break;
		}
		vma=vma->vm_next;
	}
	
	if (!entry) return ERR_PTR(-ESRCH);
	
	buf=__getname();
	if (IS_ERR(buf)) return buf;
	
	res=resolve_path(entry, buf, PAGE_SIZE);
	if (IS_ERR(res)) {
		putname(buf);
		printk(KERN_ERR
		       NAME ": Unable to resolve executable name");
	}

	return res;
} /* get_exename */

int ch_root(const char * dir)
{
	struct dentry * dentry, * tmp;
	struct super_block * rootsb;
	
	rootsb=get_super(ROOT_DEV);
	if (IS_ERR(rootsb)) return PTR_ERR(rootsb);
	
	tmp=current->fs->root;
	current->fs->root=rootsb->s_root;
	
	dentry=lookup_dentry(dir, NULL, LOOKUP_DIRECTORY);
	if (IS_ERR(dentry)) {
		current->fs->root=tmp;
		return PTR_ERR(dentry);
	}

	/* setting new root and current working directory */
	current->fs->root=dentry;
	dput(tmp);
	tmp=current->fs->pwd;
	current->fs->pwd=dget(dentry);
	dput(tmp);
	
	return 0;
} /* ch_root */

int check_chroot(void)
{
	if (IS_ROOT(current->fs->root))
		return 0;
	return -EPERM;
} /* check_chroot */
