/*
 * Copyright (C) 2001,2002 Wojciech Purczynski <cliph@isec.pl>
 *
 * THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
 * IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
 * THE FREE SOFTWARE  FOUNDATION;  EITHER  VERSION 2 OF THE LICENSE, OR
 * (AT YOUR OPTION) ANY LATER VERSION.
 *
 * THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 * BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED  WARRANTY OF
 * MERCHANTABILITY OR FITNESS  FOR A PARTICULAR  PURPOSE.  SEE THE
 * GNU GENERAL PUBLIC LICENSE FOR MORE DETAILS.
 */

#include "module.h"

#include <linux/fs.h>
#include <linux/dcache.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <asm/namei.h>

#include "config.h"
#include "fs.h"

char *resolve_path(struct dentry *entry, struct vfsmount * mnt,
                   char *buffer, int buflen)
{
	struct dentry * root;
	struct vfsmount * rootmnt;
	char * res;
	
	root=dget(entry);
	rootmnt=mntget(mnt);
	while (follow_up(&rootmnt, &root));
	
	spin_lock(&dcache_lock);
	res=__d_path(entry, mnt, rootmnt->mnt_root, rootmnt, buffer, buflen);
	spin_unlock(&dcache_lock);
	
	dput(root);
	mntput(rootmnt);
	
	return res;
} /* resolve_path */

/* NOTE: filename is from user-space memory */
char *get_realpath(const char * filename)
{
	char * buf, * res;
	struct nameidata nd;
	int error=0;

	buf=getname(filename);
	if (IS_ERR(buf)) return buf;
	
	if (path_init(buf, LOOKUP_FOLLOW|LOOKUP_POSITIVE, &nd))
		error=path_walk(buf, &nd);

	putname(buf);

	if (error) return ERR_PTR(error);
	
	buf=__getname();
	if (IS_ERR(buf)) {
		path_release(&nd);
		return buf;
	}
	
	res=resolve_path(nd.dentry, nd.mnt, buf, PAGE_SIZE);
	if (IS_ERR(res)) {
		putname(buf);
		printk(KERN_ERR
		       NAME ": Unable to resolve realpath name");
	}
	
	path_release(&nd);
	return res;
}

/*
 * Name got with get_exename must be released with put_exename()
 */
char *get_exename(void)
{
	struct vm_area_struct *vma;
	char *buf, *res;
	struct dentry * entry=NULL;
	struct vfsmount * mnt=NULL;

	vma=current->mm->mmap;
	while (vma) {
		if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file) {
			entry=vma->vm_file->f_dentry;
			mnt=vma->vm_file->f_vfsmnt;
			break;
		}
		vma=vma->vm_next;
	}
	
	if (!entry) return ERR_PTR(-ESRCH);
	
	buf=__getname();
	if (IS_ERR(buf)) return buf;
	
	res=resolve_path(entry, mnt, buf, PAGE_SIZE);
	if (IS_ERR(res)) {
		putname(buf);
		printk(KERN_ERR
		       NAME ": Unable to resolve executable name");
	}

	return res;
} /* get_exename */

/* 
 * We need that because kernel does not export this symbol :-(
 */
void set_fs_altroot(void)
{
	char *emul = __emul_prefix();
	struct nameidata nd;
	struct vfsmount *mnt = NULL, *oldmnt;
	struct dentry *dentry = NULL, *olddentry;
	if (emul) {
		read_lock(&current->fs->lock);
		nd.mnt = mntget(current->fs->rootmnt);
		nd.dentry = dget(current->fs->root);
		read_unlock(&current->fs->lock);
		nd.flags = LOOKUP_FOLLOW|LOOKUP_DIRECTORY|LOOKUP_POSITIVE;
		if (path_walk(emul,&nd) == 0) {
			mnt = nd.mnt;
			dentry = nd.dentry;
		}
	}
	write_lock(&current->fs->lock);
	oldmnt = current->fs->altrootmnt;
	olddentry = current->fs->altroot;
	current->fs->altrootmnt = mnt;
	current->fs->altroot = dentry;
	write_unlock(&current->fs->lock);
	if (olddentry) {
		dput(olddentry);
		mntput(oldmnt);
	}
} /* set_fs_altroot */

int ch_root(const char * dir)
{
	int error=0;
	struct nameidata nd;

	if (path_init(dir, LOOKUP_POSITIVE | LOOKUP_FOLLOW |
		      LOOKUP_DIRECTORY | LOOKUP_NOALT, &nd))
		error=path_walk(dir, &nd);
		
	if (error) return error;

	set_fs_root(current->fs, nd.mnt, nd.dentry);
	set_fs_pwd(current->fs, nd.mnt, nd.dentry);
	set_fs_altroot();
	path_release(&nd);

	return 0;
} /* ch_root */

int check_chroot(void)
{
	struct dentry * dentry = current->fs->root;
	struct vfsmount * vfsmnt = current->fs->rootmnt;

	if ((dentry == vfsmnt->mnt_root || IS_ROOT(dentry)) &&
	    vfsmnt->mnt_parent->mnt_parent == vfsmnt->mnt_parent)
		return 0;
	
	return -EPERM;
} /* check_chroot */

