/*
 * Copyright (C) 2001,2002 Wojciech Purczynski <cliph@isec.pl>
 *
 * THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
 * IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
 * THE FREE SOFTWARE  FOUNDATION;  EITHER  VERSION 2 OF THE LICENSE, OR
 * (AT YOUR OPTION) ANY LATER VERSION.
 *
 * THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 * BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED  WARRANTY OF
 * MERCHANTABILITY OR FITNESS  FOR A PARTICULAR  PURPOSE.  SEE THE
 * GNU GENERAL PUBLIC LICENSE FOR MORE DETAILS.
 */

#include "module.h"

#include <asm/uaccess.h>

#include "rule.h"
#include "ioctl.h"

static int capsel_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	int error=0;
	
	if (!capable(CAP_SYS_MODULE)) return -EPERM;

	switch (cmd) {
	
	/* flushes all rules - it disables capsel */
	case CAPSELVERSION: {
		int len;
		len=strlen(VERSION_STRING);
		if (len>CAPSELVERMAX) len=CAPSELVERMAX;
		error=copy_to_user((char*)arg, VERSION_STRING, len);
		break;
	} /* CAPSELVERSION */
	
	/* flushes all rules - it disables capsel */
	case CAPSELFLUSH:
		rule_free_all();
		printk(KERN_WARNING NAME ": configuration removed\n");
		break;
	
	/* creates new empty list of rules */
	case CAPSELNEW:
		rule_free_list(&new_rules);
		break;
	
	/* adds rule to the list */
	case CAPSELADD: {
		rule_t * rule;
		rule=rule_copy_from_user((rule_t*)arg);
		if (IS_ERR(rule)) {
			error=PTR_ERR(rule);
			break;
		}
		write_lock(&new_rules.lock);
		if (!new_rules.first_rule) new_rules.first_rule=rule;
		if (new_rules.last_rule) new_rules.last_rule->next=rule;
		new_rules.last_rule=rule;
		new_rules.size+=rule->size;
		new_rules.count++;
		write_unlock(&new_rules.lock);
		break;
	} /* CAPSELADD */
	
	/* adds default rule */
	case CAPSELADDDEF: {
		rule_t * rule;
		rule=rule_copy_from_user((rule_t*)arg);
		if (IS_ERR(rule)) {
			error=PTR_ERR(rule);
			break;
		}
		
		if (rule->executables) {
			rule_free(rule);
			return -EINVAL;
		}
		
		write_lock(&new_rules.lock);
		if (new_rules.default_rule) {
			new_rules.size-=new_rules.default_rule->size;
			new_rules.count--;
			rule_free(new_rules.default_rule);
		}
		new_rules.default_rule=rule;
		new_rules.size+=rule->size;
		new_rules.count++;
		write_unlock(&new_rules.lock);
		break;
	} /* CAPSELDEFAULT */

	/* install new list */
	case CAPSELINSTALL: {
		rule_list_t tmp;
		write_lock(&rules.lock);
		write_lock(&new_rules.lock);
		
		tmp=rules;
		rules=new_rules;
		new_rules=tmp;
		
		write_unlock(&new_rules.lock);
		
		printk(KERN_WARNING NAME ": new configuration "
					 "(rules=%u size=%u)\n",
					 rules.count, rules.size);
		
		write_unlock(&rules.lock);
		
		rule_free_list(&new_rules);
		
		break;
	} /* CAPSELINSTALL */
			    
	case CAPSELSTATUS: {
		capsel_status_t status;

		read_lock(&rules.lock);
		status.count = rules.count;
		status.size = rules.size;
		status.default_rule = rules.default_rule ? 1 : 0;
		read_unlock(&rules.lock);

		error=copy_to_user((void*)arg, &status, sizeof(capsel_status_t));

		break;

	} /* CAPSELSHCONF */
	
	default:
		error=-EINVAL;
	}
	
	return error;
}

static atomic_t opens=ATOMIC_INIT(0);

static int capsel_open(struct inode *inode, struct file *file)
{
#ifndef KERNEL_2_4
	MOD_INC_USE_COUNT;
#endif
	(void)inode;
	(void)file;
	atomic_inc(&opens);
	if (atomic_read(&opens)>1) {
		atomic_dec(&opens);
#ifndef KERNEL_2_4
		MOD_DEC_USE_COUNT;
#endif
		return -EBUSY;
	}
	return 0;
}

static int capsel_release(struct inode *inode, struct file *file)
{
	(void)inode;
	(void)file;
	atomic_dec(&opens);
#ifndef KERNEL_2_4
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

static struct file_operations capsel_fops = {
#ifdef KERNEL_2_4
	owner:		THIS_MODULE,
#endif
	ioctl:		capsel_ioctl,
	open:		capsel_open,
	release:	capsel_release,
};

int capsel_ioctl_init(void)
{
	int error;
#ifdef CONFIG_DEVFS_FS
	error=devfs_register_chrdev(CAPSEL_MAJOR, CAPSEL_DEV, &capsel_fops);
#else
	error=register_chrdev(CAPSEL_MAJOR, CAPSEL_DEV, &capsel_fops);
#endif
	if (error) printk(KERN_ERR NAME 
			  " unable to register character device %i\n",
			  CAPSEL_MAJOR);
	return error;
}

void capsel_ioctl_cleanup(void)
{
#ifdef CONFIG_DEVFS_FS
	(void)devfs_unregister_chrdev(CAPSEL_MAJOR, CAPSEL_DEV);
#else
	(void)unregister_chrdev(CAPSEL_MAJOR, CAPSEL_DEV);
#endif
}
