/*
 * Copyright (C) 2001,2002 Wojciech Purczynski <cliph@isec.pl>
 *
 * THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
 * IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
 * THE FREE SOFTWARE  FOUNDATION;  EITHER  VERSION 2 OF THE LICENSE, OR
 * (AT YOUR OPTION) ANY LATER VERSION.
 *
 * THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 * BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED  WARRANTY OF
 * MERCHANTABILITY OR FITNESS  FOR A PARTICULAR  PURPOSE.  SEE THE
 * GNU GENERAL PUBLIC LICENSE FOR MORE DETAILS.
 */

#ifdef __KERNEL__
#include "module.h"
#include <asm/uaccess.h>
#else
#include <stdio.h>
#endif /* __KERNEL__ */

#include "config.h"
#include "rule.h"
#include "lib.h"

#ifdef __KERNEL__ /* VERY LARGE BLOCK */

#include "fs.h"

rule_list_t rules=EMPTY_RULE_LIST;
rule_list_t new_rules=EMPTY_RULE_LIST;
const rule_t default_rule=RULE_DEFAULT;

/* Main function that allows to find matching rule for current process */
const rule_t * rule_find(void)
{
	const rule_t * rule;
	char * exe;
	
	exe=get_exename();
	if (IS_ERR(exe)) return (rule_t*)exe;
	
	for (rule=rules.first_rule; rule; rule=rule->next) {
		if (rule->name && strcmp(rule->name, exe)) continue;
		/* checking other options if paths are the same */
		if (
			/* checking UID/EUID */
			( !rule->check_uid || UID==rule->uid ) &&
			( !rule->check_euid || EUID==rule->euid ) &&
			/* checking GID/EGID */
			( !rule->check_gid || GID==rule->gid ||
			  rule->check_groups ) &&
			( !rule->check_egid || EGID==rule->egid ) &&
			/* checking supplemental group IDs */
		    	( !rule->check_gid || !rule->check_groups ||
			  in_group_p(rule->gid) )
		) break;
	}
	if (!rule) {
		if (rules.default_rule)
			rule=rules.default_rule;
		else
			rule=&default_rule;
	}
	
	put_exename(exe);
	return rule;
} /* rule_find */

int rule_change_rootdir(const rule_t * rule)
{
	int error;

	if (!rule->root_dir) return 0;

	error=ch_root(rule->root_dir);
	log(rule, LOG_DEBUG, "root directory set to %.256s", rule->root_dir);

	if (error)
		log(rule, LOG_SECURITY, "can't set root directory to %.256s",
		    rule->root_dir);

	return error;
} /* rule_change_rootdir */

/* Checking if process is allowed to use execve */
int rule_check_execve(const rule_t * rule, const char *filename)
{
	struct executable_t *exec;
	
	/* there is no exec option in default rule */
	if (rule==rules.default_rule) return 0;
	
	if (rule->check_execve==EXEC_LIST) {
		for (exec=rule->executables; exec; exec=exec->next)
			if (!strcmp(exec->name, filename)) break; /* found */

		if (exec) { /* found */
			if (rule->perm_execve==EXEC_DENY)
				return -EPERM;
			else
				return 0;
		} else { /* not found */
			if (rule->perm_execve==EXEC_ALLOW)
				return -EPERM;
			else
				return 0;
		}
	} else if (rule->check_execve==EXEC_ALL)
		return rule->perm_execve==EXEC_ALLOW?0:-EPERM;
	
	return 0;
} /* rule_check_execve */

void rule_set_caps(const rule_t * rule)
{
	const capabilities_t * caps=&rule->caps;

	/* effective */
	current->cap_effective=cap_combine(
		current->cap_effective, 
		caps->effective.raise);
	current->cap_effective=cap_drop(
		current->cap_effective, 
		caps->effective.lower);
	
	/* permitted */
	current->cap_permitted=cap_combine(
		current->cap_permitted, 
		caps->permitted.raise);
	current->cap_permitted=cap_drop(
		current->cap_permitted, 
		caps->permitted.lower);

	/* inheritable */	
	current->cap_inheritable=cap_combine(
		current->cap_inheritable, 
		caps->inheritable.raise);
	current->cap_inheritable=cap_drop(
		current->cap_inheritable, 
		caps->inheritable.lower);
} /* rule_set_caps */

char * strdup_from_user(char * from)
{
	int len;
	char * to;
	
	len=strnlen_user(from, PATH_MAX+1);
	if (len==0) return ERR_PTR(-EFAULT);
	if (len>PATH_MAX+1) return ERR_PTR(-ENAMETOOLONG);
	
	to=mmalloc(len);
	if (!to) return ERR_PTR(-ENOMEM);
	
	strncpy_from_user(to, from, len);
	return to;
}

/* copies rule and its members from user memory */
rule_t * rule_copy_from_user(const rule_t * arg)
{
	rule_t * rule;
	int error;
	executable_t * uexec;
	char * ptr;
	rule_t user;

	if (copy_from_user(&user, arg, sizeof(rule_t)))
		return ERR_PTR(-EFAULT);

	rule=mmalloc(sizeof(rule_t));
	if (!rule) return ERR_PTR(-ENOMEM);
	
	*rule=default_rule;
	
	if (user.name) {
		ptr=strdup_from_user(user.name);
		if (IS_ERR(ptr)) {
			error=PTR_ERR(ptr);
			goto err;
		}
		rule->name=ptr;
		rule->size+=strlen(ptr)+1;
	} else rule->name=NULL;

	if (user.root_dir) {
		ptr=strdup_from_user(user.root_dir);
		if (IS_ERR(ptr)) {
			error=PTR_ERR(ptr);
			goto err;
		}
		rule->root_dir=ptr;
		rule->size+=strlen(ptr)+1;
	} else rule->root_dir=NULL;
	
	rule->check_uid		= user.check_uid;
	rule->check_euid	= user.check_euid;
	rule->check_gid		= user.check_gid;
	rule->check_egid	= user.check_egid;
	rule->uid		= user.uid;
	rule->euid		= user.euid;
	rule->gid		= user.gid;
	rule->egid		= user.egid;
	rule->check_groups	= user.check_groups;
	rule->check_execve	= user.check_execve;
	rule->perm_execve	= user.perm_execve;
	rule->check_chroot	= user.check_chroot;
	rule->check_setuid	= user.check_setuid;
	rule->clear_environ	= user.clear_environ;
	rule->log_level		= user.log_level;
	rule->caps		= user.caps;
	rule->uidcaps		= user.uidcaps;
	rule->bind_tcp		= user.bind_tcp;
	rule->bind_udp		= user.bind_udp;
		
	for (uexec=user.executables; uexec;) {
		executable_t * exec;
		char * name;
		
		if (get_user(name, &uexec->name)) {
			error=-EFAULT;
			goto err;
		}

		exec=mmalloc(sizeof(executable_t));
		if (!exec) {
			error=-ENOMEM;
			goto err;
		}

		exec->next=rule->executables;
		rule->executables=exec;
		rule->size+=sizeof(executable_t);

		exec->name=NULL;

		if (!name) {
			error=-EINVAL;
			goto err;
		}
		
		ptr=strdup_from_user(name);
		if (IS_ERR(ptr)) {
			error=PTR_ERR(ptr);
			goto err;
		}
		exec->name=ptr;
		rule->size+=strlen(ptr)+1;
		
		/* uexec=uexec->next */
		if (get_user(uexec, &uexec->next)) {
			error=-EFAULT;
			goto err;
		}
	}
	
	return rule;
	
err:
	rule_free(rule);
	return ERR_PTR(error);
}

/* flushes all rules */
void rule_free_all(void)
{
	rule_free_list(&rules);
	rule_free_list(&new_rules);
} /* rule_free_all */

#endif /* __KERNEL__ */

/* frees rule and all its members */
void rule_free(rule_t * rule)
{
	executable_t * exec, * nxtexec;
	
	if (rule->name) mfree(rule->name);
	
	if (rule->root_dir) mfree(rule->root_dir);

	for (exec=rule->executables; exec; exec=nxtexec) {
		nxtexec=exec->next;
		if (exec->name)	mfree(exec->name);
		mfree(exec);
	}
	
	mfree(rule);
}

/* flushes list of rules - it locks the list */
void rule_free_list(rule_list_t * list)
{
	rule_t * rule, * nxtrule;

#ifdef __KERNEL__	
	write_lock(&list->lock);
#endif	

	/* Freeing memory */
	for (rule=list->first_rule; rule; rule=nxtrule) {
		nxtrule=rule->next;
		list->size-=rule->size;
		list->count--;
		rule_free(rule);
	}
	
	if (list->default_rule) {
		list->size-=list->default_rule->size;
		list->count--;
		rule_free(list->default_rule);
	}

	list->first_rule=NULL;
	list->last_rule=NULL;
	list->default_rule=NULL;
	if (list->count!=0 || list->size!=0) {
#ifdef __KERNEL__
		printk(KERN_WARNING 
#else
		fprintf(stderr, 
#endif
			NAME ": memory leak list->count=%u list->size=%u\n",
			list->count, list->size);
	}
	
#ifdef __KERNEL__	
	write_unlock(&list->lock);
#endif
}
