/*
 * Copyright (C) 1999-2002 Wojciech Purczynski <cliph@isec.pl>
 *
 * THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
 * IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
 * THE FREE SOFTWARE  FOUNDATION;  EITHER  VERSION 2 OF THE LICENSE, OR
 * (AT YOUR OPTION) ANY LATER VERSION.
 *
 * THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 * BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED  WARRANTY OF
 * MERCHANTABILITY OR FITNESS  FOR A PARTICULAR  PURPOSE.  SEE THE
 * GNU GENERAL PUBLIC LICENSE FOR MORE DETAILS.
 */

#include "module.h"
#include <linux/smp.h>
#include <linux/capability.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <linux/fs.h>
#include <linux/ptrace.h>
#include <linux/sysctl.h>
#include <linux/socket.h>
#include <linux/file.h>
#include <net/sock.h>

#include "config.h"
#include "sys.h"
#include "rule.h"
#include "fs.h"
#include "lib.h"

/* pointers to store original syscall handlers */
int (*old_sys_ptrace)(long, long, long, long);
rettype (*old_sys_sysctl)(struct __sysctl_args *);
rettype (*old_sys_socketcall)(int, unsigned long *);
rettype (*old_sys_chroot)(const char *);
rettype (*old_sys_setreuid)(uid_t, uid_t);
rettype (*old_sys_setuid)(uid_t);
#ifdef CONFIG_UID16
rettype (*old_sys_setreuid16)(uid_t, uid_t);
rettype (*old_sys_setuid16)(uid_t);
#endif /* CONFIG_UID16 */

extern void * sys_call_table[];

void install_hook(hook_t * hook)
{
	while (hook->number) {
		*hook->old=sys_call_table[hook->number];
		sys_call_table[hook->number]=hook->new;
		hook++;
	}
}

void uninstall_hook(hook_t * hook)
{
	while (hook->number) {
		sys_call_table[hook->number]=*hook->old;
		hook++;
	}
}

/* NEW SYSCALLS */

/* ptrace restricted to privileged users only */
asmlinkage int new_sys_ptrace(long request, long pid, long addr, long data)
{
	int error;
	if (!capable(CAP_SYS_PTRACE)) return -EPERM;
	MOD_INC_USE_COUNT;
	error=old_sys_ptrace(request, pid, addr, data);
	MOD_DEC_USE_COUNT;
	return error;
}

/* sysctl restricted to privileged users only */
asmlinkage rettype new_sys_sysctl(struct __sysctl_args * args)
{
	int error;
	if (!capable(CAP_SYS_ADMIN)) return -EPERM;
	MOD_INC_USE_COUNT;
	error=old_sys_sysctl(args);
	MOD_DEC_USE_COUNT;
	return error;
}

/* another unexported kernel function :( */
struct socket * sockfd_lookup(int fd, int * error)
{
	struct file * file;
	struct socket * sock;
	
	/* resolving socket from fd */
	file=fget(fd);
	if (!file) {
		*error=-EBADF;
		return NULL;
	}
	
	*error=-ENOTSOCK;
	if (!file->f_dentry->d_inode->i_sock) {
		fput(file);
		return NULL;
	}
	
	sock=&file->f_dentry->d_inode->u.socket_i;
	if (!sock) {
		fput(file);
		return NULL;
	}
	
	if (sock->file != file) {
		printk(KERN_ERR "sockfd_lookup: socket file changed!\n");
		sock->file = file;
	}
	return sock;
}

/*
 * This new socketcall restricts CAP_NET_BIND_SERVICE allowing to
 * bind only one specified TCP or UDP port.
 */
asmlinkage long new_sys_socketcall(int call, unsigned long *args)
{
	const rule_t * rule;
	kernel_cap_t old_effective=CAP_EMPTY_SET;
	int error=0;
	/* user arguments */
	int fd;
	struct sockaddr_in * sin;
	unsigned short port;
	/* socket datas */
	struct socket * sock;
	unsigned short family;
	unsigned char protocol;
	
	MOD_INC_USE_COUNT;
		
	if (!cap_raised(current->cap_effective, CAP_NET_BIND_SERVICE)) {
		/* standard behavior for unprivileged processes */
		error=old_sys_socketcall(call, args);
		MOD_DEC_USE_COUNT;
		return error;
	}
	
	lock_kernel_2_2();

	rule_lock();
	rule=rule_find();
	CHK_PTR(rule);

	if (!rule->bind_tcp && !rule->bind_udp) goto call;
	
	/*
	 * The default action is to drop process's privileges and
	 * raise them only if process really need them and is allowed
	 * to use CAP_NET_BIND_SERVICE and bind to specified TCP/UDP port.
	 */
	old_effective=current->cap_effective;
	cap_lower(current->cap_effective, CAP_NET_BIND_SERVICE);
	
	if (call!=SYS_BIND) goto call;
	
	error=get_user(fd, (int*)args);
	if (error) goto call;
	error=get_user(sin, (struct sockaddr_in**)args+1);
	if (error) goto call;
	
	/*
	 * We can't read socket family from user's argument because of
	 * security reason, we read this from kernel socket structure
	 * instead.
	 */
	sock=sockfd_lookup(fd, &error);
	if (!sock) goto call;
	family=sock->sk->family;
	protocol=sock->sk->protocol;
	fput(sock->file);

	error=get_user(port, &sin->sin_port);
	if (error) goto call;
	port=ntohs(port);

	error=0;
	if (family!=AF_INET && family!=AF_INET6) goto call;
	
	if ((protocol==IPPROTO_TCP && rule->bind_tcp && port==rule->bind_tcp) ||
	    (protocol==IPPROTO_UDP && rule->bind_udp && port==rule->bind_udp)) {
		current->cap_effective=old_effective; /* restore capability */
		log(rule, LOG_DEBUG, "is permitted to bind %i %s port",
				     (int)port,
				     protocol==IPPROTO_TCP?"TCP":"UDP");
	}

call:
	rule_unlock();
	if (error==0) error=old_sys_socketcall(call, args);
	/* restore original process's capabilities if needed */
	if (!cap_isclear(old_effective)) current->cap_effective=old_effective;
exit:
	unlock_kernel_2_2();
	MOD_DEC_USE_COUNT;
	return error;
}

/*
 * This function clears all current process's environment.
 * It should be portable.
 */
int clear_environ(void)
{
	return clear_user((void*)current->mm->env_start,
			  current->mm->env_end-current->mm->env_start);
}

asmlinkage rettype new_sys_chroot(const char * filename)
{
	rettype error=0;
	const rule_t * rule;
	char * realpath;

	MOD_INC_USE_COUNT;
	lock_kernel_2_2();

	rule_lock();
	rule=rule_find();
	CHK_PTR(rule);

	realpath=get_realpath(filename);
	CHK_PTR(realpath);
	
	if (rule->check_chroot)	error=check_chroot();
	
	if (!error)
		log(rule, LOG_DEBUG, "is changing root to %.256s", realpath);
	else
		log(rule, LOG_SECURITY,
		    "is not permitted to change root to %.256s", realpath);

	rule_unlock();
	if (!error) error=old_sys_chroot(filename);

	put_realpath(realpath);
exit:
	unlock_kernel_2_2();
	MOD_DEC_USE_COUNT;
	return error;
} /* new_sys_chroot */

asmlinkage rettype new_sys_setreuid(uid_t ruid, uid_t euid)
{
	const rule_t * rule;
	uid_t old_ruid, old_euid, old_suid;
	kernel_cap_t effective, permitted;
	rettype error;

	MOD_INC_USE_COUNT;	
	lock_kernel_2_2();
	
	rule_lock();
	rule=rule_find();
	CHK_PTR(rule);
	
	if (rule->check_setuid && (ruid==0 || euid==0) &&
	    (UID!=0 || EUID!=0 || SUID!=0)) {
		log(rule, LOG_SECURITY, "is trying to do setreuid(%u,%u)", ruid, euid);
		rule_unlock();
		error=-EPERM;
		goto exit;
	}

	effective=current->cap_effective;
	permitted=current->cap_permitted;
	old_ruid=current->uid;
	old_euid=current->euid;
	old_suid=current->suid;
	
	error=old_sys_setreuid(ruid, euid);

	/* it overrides original cap_emulate_setxuid */
	if (!error && !issecure(SECURE_NO_SETUID_FIXUP))
		new_cap_emulate_setxuid(rule, effective, permitted,
					old_ruid, old_euid, old_suid);
	
	rule_unlock();

exit:	
	unlock_kernel_2_2();
	MOD_DEC_USE_COUNT;
	return error;
} /* new_sys_setreuid */

asmlinkage rettype new_sys_setuid(uid_t uid)
{
	const rule_t * rule;
	uid_t old_ruid, old_euid, old_suid;
	kernel_cap_t effective, permitted;
	rettype error;
	
	MOD_INC_USE_COUNT;
	lock_kernel_2_2();
	
	rule_lock();
	rule=rule_find();
	CHK_PTR(rule);
	
	if (rule->check_setuid && uid==0 &&
	    (UID!=0 || EUID!=0 || SUID!=0)) {
		log(rule, LOG_SECURITY, "is trying to do setuid(%u)", uid);
		rule_unlock();
		error=-EPERM;
		goto exit;
	}

	effective=current->cap_effective;
	permitted=current->cap_permitted;
	old_ruid=current->uid;
	old_euid=current->euid;
	old_suid=current->suid;
	
	error=old_sys_setuid(uid);

	/* it overrides original cap_emulate_setxuid */
	if (!error && !issecure(SECURE_NO_SETUID_FIXUP))
		new_cap_emulate_setxuid(rule, effective, permitted,
					old_ruid, old_euid, old_suid);

	rule_unlock();
exit:
	unlock_kernel_2_2();
	MOD_DEC_USE_COUNT;
	return error;
} /* new_sys_setuid */

#ifdef CONFIG_UID16
#include <linux/highuid.h>
asmlinkage rettype new_sys_setreuid16(old_uid_t ruid, old_uid_t euid)
{
	return new_sys_setreuid(low2highuid(ruid), low2highuid(euid));
}

asmlinkage rettype new_sys_setuid16(old_uid_t uid)
{
	return new_sys_setuid(low2highuid(uid));
}
#endif /* CONFIG_UID16 */

hook_t sys_hooks[]={
	{ __NR_ptrace, (void**)&old_sys_ptrace, new_sys_ptrace },
	{ __NR__sysctl, (void**)&old_sys_sysctl, new_sys_sysctl },
	{ __NR_socketcall, (void**)&old_sys_socketcall, new_sys_socketcall },
	{ __NR_chroot, (void**)&old_sys_chroot, new_sys_chroot },
#ifdef CONFIG_UID16
	{ __NR_setreuid32, (void**)&old_sys_setreuid, new_sys_setreuid },
	{ __NR_setuid32, (void**)&old_sys_setuid, new_sys_setuid },
	{ __NR_setreuid, (void**)&old_sys_setreuid16, new_sys_setreuid16 },
	{ __NR_setuid, (void**)&old_sys_setuid16, new_sys_setuid16 },
#else /* CONFIG_UID16 */
	{ __NR_setreuid, (void**)&old_sys_setreuid, new_sys_setreuid },
	{ __NR_setuid, (void**)&old_sys_setuid, new_sys_setuid },
#endif /* CONFIG_UID16 */
	{ 0, NULL, NULL }
};
