/*
 * Copyright (C) 1999-2002 Wojciech Purczynski <cliph@isec.pl>
 *
 * THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
 * IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
 * THE FREE SOFTWARE  FOUNDATION;  EITHER  VERSION 2 OF THE LICENSE, OR
 * (AT YOUR OPTION) ANY LATER VERSION.
 *
 * THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 * BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED  WARRANTY OF
 * MERCHANTABILITY OR FITNESS  FOR A PARTICULAR  PURPOSE.  SEE THE
 * GNU GENERAL PUBLIC LICENSE FOR MORE DETAILS.
 */

#include "config.h"
#include "conf.h"
#include "file.h"
#include "rule.h"
#include "lib.h"

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <netinet/in.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>

rule_list_t rules=EMPTY_RULE_LIST;
rule_t * default_rule_backup;	/* default rule separate for each file */

typedef struct {
	char * name;
	kernel_cap_t value;
} cap_name_t;

/* Definition of capabilities names and their bits value */
cap_name_t cap_names[];

#define IS(str, val) (!strcmp((str), (val)))
#define ISNT(str, val) (strcmp((str), (val)))
#define IS_STRING(str) (1)
#define IS_TRUE(str) (IS((str), "true") || IS((str), "yes") || IS((str), "on"))
#define IS_FALSE(str) (IS((str), "false") || IS((str), "no") || IS((str), "off"))
#define IS_BOOL(str) (IS_TRUE(str) || IS_FALSE(str))
#define ISNT_BOOL(str) (!(IS_BOOL(str)))
#define IS_INT(str) (isdigit(*(str)))
#define ISNT_INT(str) (!(IS_INT(str)))

/* XXX: make it to be a macro with formatted messages */
int __file_error_at(char * file, int line, char * msg)
{
	fprintf(stderr, NAME ": ");
	if (file) fprintf(stderr, "%s: ", file);
	fprintf(stderr, "%s", msg);
	if (line>0)
		fprintf(stderr, " at line %i\n", file_line_no());
	else
		fprintf(stderr, "\n");
	return -1;
}

#define file_error_at(msg) __file_error_at(file_name(), file_line_no(), msg)
#define file_error(msg) __file_error_at(file_name(), 0, msg)

/* This function reads one word from configuration file. */
char * read_word(void)
{
	char * word;
	char chr;
	
	for (;;) {
		word=file_read_word();
		if (!word) return NULL;
		
		if (!strncmp(word, "//", 2) || *word=='#') {
			do chr=file_get_char();
			while (chr>0 && chr!=0x0a);
			if (chr<=0) return NULL;
		} else if (!strncmp(word, "/*",2)) {
			for (;;) {
				int len;
				word=file_read_word();
				if (!word) return NULL;
				len=strlen(word);
				if (len>=2 && !strcmp(word+len-2, "*/")) break;
			}
		} else break;
	}
	return word;
}

/* 
 * These macros convert strings to proper types.
 * Note, that these macros call file_error() and immediately returns from
 * the function with negative value on any error detected.
 */
#define S_TO_STRING(var, str)						\
	do {								\
		(var)=strdup(str);					\
		if (!(var))						\
			return file_error("Out of memory");		\
	} while (0)

#define S_TO_BOOL(var, str)						\
	do {								\
		(var)=(IS_TRUE(str))?TRUE:FALSE;			\
	} while (0)	

#define S_TO_INT(var, str)						\
	do {								\
		(var)=strtoul((str), NULL, 10);				\
	} while(0)

#define READ_WORD(str)	 						\
	do {								\
		(str)=read_word();					\
		if (!(str))						\
			return file_error_at("Unexpected end of file");	\
	} while(0)

/* Generic READ_xxx macro */
#define READ_TYPE(type, desc, var) 					\
	do {								\
		char * str;						\
		READ_WORD(str);						\
		if (!(IS_##type(str)))					\
			return file_error_at(desc " expected");		\
		S_TO_##type((var), (str));					\
	} while(0)

/* These macros are usefull to parse options and their values */
#define READ_STRING(var) READ_TYPE(STRING, "String value", (var))
#define READ_BOOL(var)	READ_TYPE(BOOL, "Boolean value", (var))
#define READ_INT(var) READ_TYPE(INT, "Integer value", (var))

#define EXPECT(val)							\
	do {								\
		char * str;						\
		READ_WORD(str);						\
		if (!(IS_STRING(str)) || ISNT((str), (val)))		\
			return file_error_at("`"val"' expected"); 	\
	} while(0)

#define LVL_LOWER	0
#define LVL_RAISE	1
#define LVL_SET		2
#define LVL_MASK	3

int read_rule(const char *);

/*
 * This function creates new copy of a rule.
 * Restrictions:
 *	name is always NULL (no name)
 *	executables is always NULL (empty list)
 *	root_dir is always copied
 */
rule_t * copy_rule(rule_t * from)
{
	rule_t * rule;
	executable_t * exec;
	
	rule=malloc(sizeof(rule_t));
	if (!rule) {
		file_error("Out of memory for rule_t");
		return NULL;
	}

	*rule=*from;
	rule->next=NULL;
	
	/* creating empty name */
	if (rule->name) {
		rule->size-=strlen(rule->name)+1;
		rule->name=NULL;
	}

	/* clearing list of executables - it will be copied */
	for (exec=rule->executables; exec; exec=exec->next) {
		rule->size-=sizeof(executable_t);
		if (exec->name) rule->size-=strlen(exec->name)+1;
	}
	rule->executables=NULL;
	
	/* copying root_dir */
	if (rule->root_dir) {
		rule->size-=strlen(rule->root_dir)+1;
		rule->root_dir=strdup(rule->root_dir);
		if (!rule->root_dir) {
			rule_free(rule);
			file_error("Out of memory for root_dir");
			return NULL;
		}
		rule->size+=strlen(rule->root_dir)+1;
	}
	
	return rule;
}

int read_conf(const char * file, int global)
{
	char * str;
	int error;

	printf("Parsing: %s\n", file);

	error=file_open(file);
	if (error<0) {
		file_error("Can't open configuration file");
		goto exit;
	}
	
	/*
	 * Global default rule from /etc/capsel.conf is
	 * left unchanged and we use local copy for each
	 * non-global configuration file.
	 */
	if (!global) {
		if (rules.default_rule) {
			default_rule_backup=copy_rule(rules.default_rule);
			if (!default_rule_backup) goto exit;
		} else default_rule_backup=NULL;
	}

	/* parsing the whole configuration file */
	while ( (str=read_word())!=NULL ) {
		if (IS(str, "eof")) break;
		else if (IS(str, "end")) break;
		else {
			error=read_rule(str);
			if (error) break;
		}
	}

exit:
	file_close();
		
	if (!global) {
		if (rules.default_rule) {
			if (!default_rule_backup) {
				rules.size-=rules.default_rule->size;
				rules.count--;
			}
			rule_free(rules.default_rule);
		}
		rules.default_rule=default_rule_backup;
	}
	
	return error;
}

int read_options(rule_t*);
rule_t * last_rule=NULL;

#define IS_DEFAULT(filename) (IS(filename, "default"))

int read_rule(const char * filename)
{
	char * str;
	rule_t * rule=NULL;

	/* Allocating new rule */
	if (IS(filename, "default")) {
		rule=rules.default_rule;
	}
	
	if (!rule) {
		if (rules.default_rule)	{
			rule=copy_rule(rules.default_rule);
		} else {
			rule=copy_rule(&RULE_DEFAULT);
		}
		
		if (!rule)
			return file_error_at("Out of memory for rule_t");

		if (IS_DEFAULT(filename)) {
			rules.default_rule=rule;
		} else {
			if (!rules.first_rule) rules.first_rule=rule;
			if (rules.last_rule) rules.last_rule->next=rule;
			rules.last_rule=rule;
		}
		
		rules.size+=rule->size;
		rules.count++;
	}

	if (*filename=='/') {
		struct stat st;
		
		if (lstat(filename, &st))
			return file_error_at("Can't stat file");
		if (!S_ISREG(st.st_mode))
			return file_error_at("Not a regular file");
		
		rule->name=strdup(filename);
		if (!rule->name) 
			return file_error_at("Out of memory for filename");
		
		rule->size+=strlen(rule->name)+1;
		rules.size+=strlen(rule->name)+1;
		
	} else if (!IS_DEFAULT(filename) && ISNT(filename, "any"))
		return file_error_at("Absolute pathname expected");

	EXPECT("{");

	return read_options(rule);
}

int read_cap_opt(capabilities_t*, int, const capabilities_t *);
int read_exec_opt(rule_t*);

/*
 * Functions that parse rule's option.
 * Each parser takes two arguments (the rule and option structure pointer)
 * and returns 0 if it succeded parsing the option. In case of error it
 * returns -1.
 */

typedef struct option_t {
	char * name;					/* name of option */
	int (*parser)(struct option_t *, rule_t*);	/* parser */
} option_t;

int parse_uid(option_t * option, rule_t * rule)
{
	int check_uid;
	uid_t uid;
	char * val;

	READ_WORD(val);
	if (IS(val, "any")) {
		check_uid=FALSE;
		uid=0;
	} else {
		struct passwd * pw;
		if (IS_INT(val)) {
			S_TO_INT(uid, val);
			if (uid<0) return file_error_at("Invalid UID value");
			pw=getpwuid(uid);
			if (!pw) return file_error_at("UID not found");
		} else {
			pw=getpwnam(val);
			if (!pw) return file_error_at("User not found");
		}
		check_uid=TRUE;
		uid=pw->pw_uid;
	}
	if (IS(option->name, "uid") || IS(option->name, "user")) {
		rule->check_uid=check_uid;
		rule->uid=uid;
	} else { /* euid */
		rule->check_euid=check_uid;
		rule->euid=uid;
	}
	return 0;
}

int parse_gid(option_t * option, rule_t * rule)
{
	int check_gid;
	gid_t gid;
	char * val;
	
	READ_WORD(val);
	if (IS(val, "any")) {
		check_gid=FALSE;
		gid=0;
	} else {
		struct group * gr;
		if (IS_INT(val)) {
			S_TO_INT(gid, val);
			if (gid<0) return file_error_at("Invalid GID value");
			gr=getgrgid(gid);
			if (!gr) return file_error_at("GID not found");
		} else {
			gr=getgrnam(val);
			if (!gr) return file_error_at("Group not found");
		}
		check_gid=TRUE;
		gid=gr->gr_gid;
	}
	if (IS(option->name, "gid") || IS(option->name, "group")) {
		rule->check_gid=check_gid;
		rule->gid=gid;
	} else { /* euid */
		rule->check_egid=check_gid;
		rule->egid=gid;
	}
	return 0;
}

option_t options[]={
	{ "uid", parse_uid },
	{ "user", parse_uid },
	{ "euid", parse_uid },
	{ "gid", parse_gid },
	{ "group", parse_gid },
	{ "egid", parse_gid },
	{ NULL, NULL }
};

int read_options(rule_t * rule)
{
	char * str;
	int error;

	for (;;) {
		option_t * option;
		
		READ_WORD(str);
		if (IS(str, "}")) break; /* end of rule */
		
		/* some of options are parsed by separate functions... */
		for (option=options; option->name; option++) {
			if (IS(str, option->name)) {
				if (option->parser(option, rule)) return -1;
				else break;
			}
		}
		
		if (option->name) continue; /* already parsed */
		
		/* ...and some are not yet */
		if (IS(str, "groups")) {
			READ_BOOL(rule->check_groups);
		} else if (IS(str, "chroot") || IS(str, "chkchroot")) {
			READ_BOOL(rule->check_chroot);
		} else if (IS(str, "clrenv")) {
			READ_BOOL(rule->clear_environ);
		} else if (IS(str, "chksetuid")) {
			READ_BOOL(rule->check_setuid);
		} else if (IS(str, "log")) {
			READ_WORD(str);
			if (IS(str, "nothing"))
				rule->log_level=LOG_NOTHING;
			else if (IS(str, "debug"))
				rule->log_level=LOG_DEBUG;
			else if (IS(str, "security"))
				rule->log_level=LOG_SECURITY;
			else return file_error_at("Syntax error");
		} else if (IS(str, "cap")) {
			error=read_cap_opt(&rule->caps, LVL_SET, &CAP_DEFAULT);
			if (error) return error;
		} else if (IS(str, "uidcap")) {
			error=read_cap_opt(&rule->uidcaps, LVL_MASK,
					   &UIDCAP_DEFAULT);
			if (error) return error;
		} else if (IS(str, "bindtcp")) {
			int port;
			READ_INT(port);
			if (port<0 || port>=IPPORT_RESERVED)
				return file_error_at("Invalid port number");
			rule->bind_tcp=(unsigned short)port;
		} else if (IS(str, "bindudp")) {
			int port;
			READ_INT(port);
			if (port<0 || port>=IPPORT_RESERVED)
				return file_error_at("Invalid port number");
			rule->bind_udp=(unsigned short)port;
		} else if (IS(str, "exec")) {
			if (rule==rules.default_rule)
				return file_error_at("Invalid option in default rule");
			if (rule->executables)
				return file_error_at("Duplicate exec option");
			error=read_exec_opt(rule);
			if (error) return error;
		} else if (IS(str, "root") || IS(str, "rootdir")) {
			READ_WORD(str);
			if (rule->root_dir)
				mfree(rule->root_dir);
			if (ISNT(str, "default")) 
				S_TO_STRING(rule->root_dir, str);
		} else if (IS(str, "ignore")) {
			int i;
			READ_BOOL(i);
		} else return file_error_at("Unknown option");
	}
	return 0;
}

int read_exec_opt(rule_t * rule)
{
	char * str;
	
	READ_WORD(str);
	
	if (IS(str, "allow")) {
		rule->perm_execve=EXEC_ALLOW;
		READ_WORD(str);
	} else if (IS(str, "deny")) {
		rule->perm_execve=EXEC_DENY;
		READ_WORD(str);
	}
	
	if (IS(str, "all")) {
		rule->check_execve=EXEC_ALL;
		return 0;
	}
	
	if (ISNT(str, "{"))
		return file_error_at("`{' expected");

	rule->check_execve=EXEC_LIST;
	for (;;) {
		executable_t * exec;
		READ_WORD(str);
		if (IS(str, "}")) break;
		if (*str!='/')
			return file_error_at("Absolute pathname expected");
		exec=(executable_t*)malloc(sizeof(executable_t));
		if (!exec) return file_error_at("Out of memory for executable_t");

		exec->name=NULL;
		exec->next=rule->executables;
		rule->executables=exec;
	
		S_TO_STRING(exec->name, str);
	} 

	return 0;
}

int read_cap_set(kernel_cap_t*);
void rule_set_cap(cap_mask_t *, kernel_cap_t, int);

int read_cap_opt(capabilities_t * caps, int level, const capabilities_t * def)
{
	char * str;
	int error;
	int type;
	kernel_cap_t capset;

	type=0;
	for (;;) {
		READ_WORD(str);
		if (IS(str, "effective"))
			type|=CAP_EFFECTIVE;
		else if (IS(str, "permitted"))
			type|=CAP_PERMITTED;
		else if (IS(str, "inheritable"))
			type|=CAP_INHERITABLE;
		else break;
	}
	
	if (type==0) type=CAP_EFFECTIVE|CAP_PERMITTED|CAP_INHERITABLE;
	
	if (IS(str, "set")) {
		level=LVL_SET;
		READ_WORD(str);
	} else if (IS(str, "raise")) {
		level=LVL_RAISE;
		READ_WORD(str);
	} else if (IS(str, "lower")) {
		level=LVL_LOWER;
		READ_WORD(str);
	} else if (IS(str, "mask")) {
		level=LVL_MASK;
		READ_WORD(str);
	} else if (IS(str, "default")) {
		if (type & CAP_EFFECTIVE)
			caps->effective=def->effective;
		if (type & CAP_PERMITTED)
			caps->permitted=def->permitted;
		if (type & CAP_INHERITABLE)
			caps->inheritable=def->inheritable;
		return 0;
	}

	if (IS(str, "empty"))
		cap_clear(capset);
	else if (IS(str, "full"))
		cap_set_full(capset);
	else if (IS(str, "{")) {
		error=read_cap_set(&capset);
		if (error) return error;
	} else return file_error_at("Syntax error");
	
	if (type & CAP_EFFECTIVE) 
		rule_set_cap(&caps->effective, capset, level);
	if (type & CAP_PERMITTED)
		rule_set_cap(&caps->permitted, capset, level);
	if (type & CAP_INHERITABLE)
		rule_set_cap(&caps->inheritable, capset, level);

	return 0;
}

int read_cap_set(kernel_cap_t * caps)
{
	char * str;
	
	cap_clear(*caps);
	
	for (;;) {
		cap_name_t * cap;
		READ_WORD(str);
		if (IS(str, "}")) break;
		for (cap=cap_names; cap->name; cap++)
			if (IS(str, cap->name)) break;
		if (cap->name) *caps=cap_combine(*caps, cap->value);
		else return file_error_at("Invalid capability name");
	}
	return 0;
}

void rule_set_cap(cap_mask_t * capmask, kernel_cap_t capval, int level)
{
	switch (level) {
	case LVL_LOWER: {
		capmask->raise=cap_drop(capmask->raise, capval);
		capmask->lower=cap_combine(capmask->lower, capval);
	} break;
	case LVL_RAISE: {
		capmask->raise=cap_combine(capmask->raise, capval);
		capmask->lower=cap_drop(capmask->lower, capval);
	} break;
	case LVL_SET: {
		capmask->raise=capval;
		capmask->lower=cap_invert(capval);
	} break;
	case LVL_MASK: {
		capmask->raise=cap_intersect(capmask->raise, capval);
		capmask->lower=cap_invert(capval);
	} break;
	}
}

cap_name_t cap_names[]={
	{ "chown", to_cap_t(CAP_TO_MASK(CAP_CHOWN)) },
	
	{ "dac_override", to_cap_t(CAP_TO_MASK(CAP_DAC_OVERRIDE)) },
	{ "dac", to_cap_t(CAP_TO_MASK(CAP_DAC_OVERRIDE)) },
	
	{ "dac_read_search", to_cap_t(CAP_TO_MASK(CAP_DAC_READ_SEARCH)) },
	{ "dacrs", to_cap_t(CAP_TO_MASK(CAP_DAC_READ_SEARCH)) },
	
	{ "fowner", to_cap_t(CAP_TO_MASK(CAP_FOWNER)) },
	
	{ "fsetid", to_cap_t(CAP_TO_MASK(CAP_FSETID)) },
	
	{ "fs_mask", to_cap_t(CAP_FS_MASK) },
	{ "fs", to_cap_t(CAP_FS_MASK) },
	
	{ "kill", to_cap_t(CAP_TO_MASK(CAP_KILL)) },
	
	{ "setgid", to_cap_t(CAP_TO_MASK(CAP_SETGID)) },
	
	{ "setuid", to_cap_t(CAP_TO_MASK(CAP_SETUID)) },
	
	{ "setpcap", to_cap_t(CAP_TO_MASK(CAP_SETPCAP)) },
	
	{ "linux_immutable", to_cap_t(CAP_TO_MASK(CAP_LINUX_IMMUTABLE)) },
	{ "immutable", to_cap_t(CAP_TO_MASK(CAP_LINUX_IMMUTABLE)) },
	
	{ "net_bind_service", to_cap_t(CAP_TO_MASK(CAP_NET_BIND_SERVICE)) },
	{ "bind", to_cap_t(CAP_TO_MASK(CAP_NET_BIND_SERVICE)) },
	
	{ "net_broadcast", to_cap_t(CAP_TO_MASK(CAP_NET_BROADCAST)) },
	{ "broadcast", to_cap_t(CAP_TO_MASK(CAP_NET_BROADCAST)) },
	
	{ "net_admin", to_cap_t(CAP_TO_MASK(CAP_NET_ADMIN)) },
	{ "netadmin", to_cap_t(CAP_TO_MASK(CAP_NET_ADMIN)) },
	{ "net", to_cap_t(CAP_TO_MASK(CAP_NET_ADMIN)) },
	
	{ "net_raw", to_cap_t(CAP_TO_MASK(CAP_NET_RAW)) },
	{ "raw", to_cap_t(CAP_TO_MASK(CAP_NET_RAW)) },
	
	{ "ipc_lock", to_cap_t(CAP_TO_MASK(CAP_IPC_LOCK)) },
	
	{ "ipc_owner", to_cap_t(CAP_TO_MASK(CAP_IPC_OWNER)) },
	{ "ipc", to_cap_t(CAP_TO_MASK(CAP_IPC_OWNER)) },
	
	{ "sys_module", to_cap_t(CAP_TO_MASK(CAP_SYS_MODULE)) },
	{ "module", to_cap_t(CAP_TO_MASK(CAP_SYS_MODULE)) },
	
	{ "sys_rawio", to_cap_t(CAP_TO_MASK(CAP_SYS_RAWIO)) },
	{ "rawio", to_cap_t(CAP_TO_MASK(CAP_SYS_RAWIO)) },
	
	{ "sys_chroot", to_cap_t(CAP_TO_MASK(CAP_SYS_CHROOT)) },
	{ "chroot", to_cap_t(CAP_TO_MASK(CAP_SYS_CHROOT)) },
	
	{ "sys_ptrace", to_cap_t(CAP_TO_MASK(CAP_SYS_PTRACE)) },
	{ "ptrace", to_cap_t(CAP_TO_MASK(CAP_SYS_PTRACE)) },
	
	{ "sys_pacct", to_cap_t(CAP_TO_MASK(CAP_SYS_PACCT)) },
	{ "pacct", to_cap_t(CAP_TO_MASK(CAP_SYS_PACCT)) },
	
	{ "sys_admin", to_cap_t(CAP_TO_MASK(CAP_SYS_ADMIN)) },
	{ "admin", to_cap_t(CAP_TO_MASK(CAP_SYS_ADMIN)) },
	
	{ "sys_boot", to_cap_t(CAP_TO_MASK(CAP_SYS_BOOT)) },
	{ "boot", to_cap_t(CAP_TO_MASK(CAP_SYS_BOOT)) },
	
	{ "sys_nice", to_cap_t(CAP_TO_MASK(CAP_SYS_NICE)) },
	{ "nice", to_cap_t(CAP_TO_MASK(CAP_SYS_NICE)) },
	
	{ "sys_resource", to_cap_t(CAP_TO_MASK(CAP_SYS_RESOURCE)) },
	{ "resource", to_cap_t(CAP_TO_MASK(CAP_SYS_RESOURCE)) },
	
	{ "sys_time", to_cap_t(CAP_TO_MASK(CAP_SYS_TIME)) },
	{ "time", to_cap_t(CAP_TO_MASK(CAP_SYS_TIME)) },
	
	{ "sys_tty_config", to_cap_t(CAP_TO_MASK(CAP_SYS_TTY_CONFIG)) },
	{ "ttyconfig", to_cap_t(CAP_TO_MASK(CAP_SYS_TTY_CONFIG)) },

#ifdef CAP_MKNOD
	{ "mknod", to_cap_t(CAP_TO_MASK(CAP_MKNOD)) },
#endif
#ifdef CAP_LEASE
	{ "lease", to_cap_t(CAP_TO_MASK(CAP_LEASE)) },
#endif

	{ NULL, CAP_EMPTY_SET }
};
